/*
 * mmap a file
 *
 * -f name	Specify test file, - for stdin.
 * -o offset	Offset starting from which file should be mapped
 * -l size	Length of the file region to me mapped.
 * -t		Truncate file to size given by -l prior to mapping. 
 * -tt		Truncate file to 0, then proceed as with -t.
 * -v		Verify any data written.
 * -r count	Repeat operation count times (-1 means forever).
 * -R		Remap file after each iteration.
 *
 * -s		sync mmapped area
 * -i		invalidate mmapped area
 * -T		Check timing of read vs. write.
 *
 * -a msec	Arrange for a timer signal to be delivered to the
 *		process every msec milliseconds.
 *
 * All numeric values (except the delay) can be specified with an
 * optional unit of k, m, or g, meaning kilo-, mega- or gigabyte.
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

/* Must be power of 2 */
#define CHUNK		64 * 1024

static long		getnum(const char *);
static void		fuse(unsigned int msec);

static void
usage(int code)
{
	printf("usage:\n"
		"[-o offset] [-l length] "
		"[-r repeat] [-iRstTv] "
		"[-a alarmintvl] [-f filename]\n");
	exit(code);
}

int
main(int argc, char **argv)
{
	const char	*fname = "junk";
	char		buffer[CHUNK], *area;
	int		length = 4096 * 4096,
			interval = 0,
			offset = 0,
			repeat = -1,
			otruncate = 0,
			osync = 0,
			overify = 0,
			oremap = 0,
			oinvalidate = 0,
			otimer = 0,
			c, fd = -1, where, left, i, j, verifying;
	long		duration, initial;
	struct timeval	t0, t1;

	while ((c = getopt(argc, argv, "a:f:hil:o:r:RstTv")) != -1) {
		switch (c) {
		case 'a':
			interval = getnum(optarg);
			break;
		case 'f':
			fname = optarg;
			break;
		case 'i':
			oinvalidate++;
			break;
		case 'l':
			length = getnum(optarg);
			break;
		case 'o':
			offset = getnum(optarg);
			break;
		case 'R':
			oremap = 1;
			break;
		case 'r':
			repeat = getnum(optarg);
			break;
		case 's':
			osync++;
			break;
		case 't':
			otruncate++;
			break;
		case 'T':
			otimer++;
			break;
		case 'v':
			overify++;
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}

	if (!strcmp(fname, "-"))
		fd = 0;

	/* init buffer */
	memcpy(buffer, "0123456789abcdefghijklmnopqrstuv", 32);
	for (i = 32; i < CHUNK; i <<= 1)
		memcpy(buffer + i, buffer, i);

	if (repeat == 0)
		repeat = 1;

	if (interval)
		fuse(interval);

	verifying = 0;
	where = offset;
	while (1) {
	again:
		if (fd < 0) {
			int	openmode = O_RDWR|O_CREAT;

			if (otruncate >= 2)
				openmode |= O_TRUNC;
			if ((fd = open(fname, openmode, 0666)) < 0) {
				perror(fname);
				exit(1);
			}
			if (otruncate)
				ftruncate(fd, offset + length);
			area = mmap(NULL, length, 
					PROT_READ|PROT_WRITE, MAP_SHARED,
					fd, offset);
			if (area == 0) {
				perror("mmap");
				exit(1);
			}
			initial = 0;
		}

		left = length;
		where = i = 0;

		gettimeofday(&t0, NULL);
		while (left) {
			j = left;
			if (i + j > CHUNK)
				j = CHUNK - i;
			if (!verifying) {
				memcpy(area + where, buffer + i, j);
			} else {
				if (memcmp(area + where, buffer + i, j)) {
					printf("Buffer mismatch in %d-%d\n",
						where, where + j);
					exit(1);
				}
			}
			if ((i += j) >= CHUNK)
				i = 0;
			where += j;
			left -= j;
		}

		if (osync || oinvalidate) {
			int	how = MS_SYNC;

			if (oinvalidate)
				how |= MS_INVALIDATE;
			if (msync(area, length, how) < 0) {
				perror("msync");
				exit(1);
			}
			if (oinvalidate > 1) {
				if (fchmod(fd, 0000) < 0
				 || fchmod(fd, 0600) < 0) {
					perror("fchmod");
					exit(1);
				}
			}
		}

		gettimeofday(&t1, NULL);
		duration = (t1.tv_sec - t0.tv_sec) * 1000
			+ (t1.tv_usec - t0.tv_usec) / 1000;
		if (verifying && initial && oinvalidate && otimer) {
			if (length > 1024 * 1024 && duration < initial / 4) {
				printf("Warning: MS_INVALIDATE doesn't "
					"seem to discard the page cache "
					"(write %ld msec read %ld msec)\n",
					initial, duration);
				exit(1);
			}
		}
		if (!initial)
			initial = duration;

		if (overify) {
			verifying = !verifying;;
			if (verifying)
				goto again;
		}

		if (repeat > 0 && --repeat == 0)
			break;
		if (oremap) {
			if (munmap(area, length) < 0) {
				perror("munmap");
				exit(1);
			}
			if (close(fd) < 0) {
				perror("close");
				exit(1);
			}
			fd = -1;
		}
	}
	return 0;
}

/*
 * interval timer
 */
static void
blip(int sig)
{
	/* printf("."); fflush(stdout); */
	return;
}

static void
fuse(unsigned int msec)
{
	struct sigaction	act;
	struct itimerval	timer;

	memset(&act, 0, sizeof(act));
	sigemptyset(&act.sa_mask);
	act.sa_handler = blip;
	sigaction(SIGALRM, &act, NULL);

	timer.it_value.tv_sec  = msec / 1000;
	timer.it_value.tv_usec = (msec % 1000) * 1000;
	timer.it_interval = timer.it_value;

	setitimer(ITIMER_REAL, &timer, NULL);
}

static long
getnum(const char *string)
{
	char	*end;
	long	val;

	val = strtol(string, &end, 0);
	if (*end == 'k' || *end == 'K')
		val *= 1024;
	else if (*end == 'm' || *end == 'M')
		val *= 1024 * 1024;
	else if (*end == 'g' || *end == 'G')
		val *= 1024 * 1024 * 1024;
	else if (*end == '\0')
		return val;
	if (*++end != '\0') {
		printf("Invalid number \"%s\"\n", string);
		exit(1);
	}

	return val;
}
